;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "nerd-icons-corfu" "20260823.1748"
  "Icons for Corfu via nerd-icons."
  '((emacs      "27.1")
    (nerd-icons "0.1.0"))
  :url "https://github.com/LuigiPiucco/nerd-icons-corfu"
  :commit "e1197d6c1db673f4ec7ee20cb2c4297f479420e7"
  :revdesc "e1197d6c1db6"
  :keywords '("convenience" "files" "icons")
  :authors '(("Luigi Sartor Piucco" . "luigipiucco@gmail.com"))
  :maintainers '(("Luigi Sartor Piucco" . "luigipiucco@gmail.com")))
